/******************************************************************************/
/*                                                                            */
/* Version 1.5                                                                */
/* Author: Arjan van Vught                 Date: October   2001               */
/*                                                                            */
/******************************************************************************/
/* Version 1.0   First release NT                                             */
/* Version 1.1   First release AIX                                            */
/* Version 1.2   Updated for MQSeries Integrator 2.0.1                        */
/* Version 1.3   First release Sun Solaris                                    */
/* Version 1.4   First release Linux (on Intel)                               */
/* Version 1.5   First release HP-UX                                          */
/******************************************************************************/
/*                                                                            */
/* Note: This plugin node implementation is not necessarily portable          */
/* to all platforms supported by MQSeries Integrator Version 2.               */
/*                                                                            */
/******************************************************************************/
/* Includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>
#include <ctype.h>

#include "FFMSG001parser.h"

CciContext *cpiCreateContext(/*@unused@*/CciParser *parser)
{
   PARSER_CONTEXT_ST *pc = NULL;

#ifdef _DEBUG  
   DBGPRINTF((OST, "->cpiCreateContext()\n"));
#endif      

   /* Allocate a pointer to the local context */
   pc = (PARSER_CONTEXT_ST *)malloc(sizeof(PARSER_CONTEXT_ST));

   if (pc) {
      /* Clear the context area */
      memset(pc, 0, sizeof(PARSER_CONTEXT_ST));
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
		CciCharCpy(pc->NextParserClassName, constFFMSG001);
#else
      wcscpy(pc->NextParserClassName, L"FFMSG001");
#endif
   }
   else {
    /* Error: Unable to allocate memory */
   }

#ifdef _DEBUG  
   DBGPRINTF((OST, "<-cpiCreateContext()\n"));
#endif      

   /* Return the pointer to the local context */
   return((CciContext *)pc);
}


void cpiDeleteContext(/*@unused@*/CciParser *parser, CciContext *c)
{
   PARSER_CONTEXT_ST *pc = (PARSER_CONTEXT_ST *)c;

#ifdef _DEBUG  
   DBGPRINTF((OST, "->cpiDeleteContext()\n"));
#endif

   free(pc);

#ifdef _DEBUG  
   DBGPRINTF((OST, "<-cpiDeleteContext()\n"));
#endif

   return;
}


/******************************************************************************/
/*                                                                            */
/* Plugin Parser Implementation Function:                                     */
/*                                                                            */
/******************************************************************************/
void cpiParseFirstChild(CciParser *parser, CciContext *context,  CciElement *element)
{
   PARSER_CONTEXT_ST* pc = context;
   int                rc;

#ifdef _DEBUG  
   DBGPRINTF((OST,"->cpiParseFirstChild() parser = 0x%x, context = 0x%x, element = 0x%x\n", parser, context, element));
#endif  

   if ((!cpiElementCompleteNext(&rc, element)) && (cpiElementType(&rc, element) == CCI_ELEMENT_TYPE_NAME)) {
      while ((!cpiElementCompleteNext(&rc, element)) && (!cpiFirstChild(&rc, element)) && (pc->iCurrentElement))
         pc->iCurrentElement = parseNextItem(parser, context, pc->iCurrentElement);
   }

#ifdef _DEBUG  
   DBGPRINTF((OST,"<-cpiParseFirstChild()\n"));
#endif

   return;
}

/******************************************************************************/
/*                                                                            */
/* Plugin Parser Implementation Function:                                     */
/*                                                                            */
/******************************************************************************/
void cpiParseLastChild(CciParser *parser, CciContext *context, CciElement *element)
{
   PARSER_CONTEXT_ST* pc = context;
   int                rc;

#ifdef _DEBUG  
   DBGPRINTF((OST, "->cpiParseLastChild() parser = 0x%x, context = 0x%x, element = 0x%x\n", parser, context, element));
#endif
  
   if ((cpiElementType(&rc, element) == CCI_ELEMENT_TYPE_NAME)) {
      while ((!cpiElementCompleteNext(&rc, element)) && (pc->iCurrentElement))
         pc->iCurrentElement = parseNextItem(parser, context, pc->iCurrentElement);
   }

#ifdef _DEBUG  
   DBGPRINTF((OST, "<-cpiParseLastChild()\n"));
#endif

  return;
}

/******************************************************************************/
/*                                                                            */
/* Plugin Parser Implementation Function:                                     */
/*                                                                            */
/******************************************************************************/
void cpiParseLeftSibling(CciParser *parser, CciContext *context, CciElement *element)
{
#ifdef _DEBUG  
   DBGPRINTF((OST, "->cpiParseLeftSibling() parser = 0x%x, context = 0x%x, element = 0x%x\n", parser, context, element));
  
   DBGPRINTF((OST, "<-cpiParseLeftSibling()\n"));
#endif

   return;
}

/******************************************************************************/
/*                                                                            */
/* Plugin Parser Implementation Function:                                     */
/*                                                                            */
/******************************************************************************/
void cpiParseRightSibling(CciParser *parser, CciContext *context, CciElement *element)
{
   PARSER_CONTEXT_ST* pc = context;
   int                rc;

#ifdef _DEBUG  
   DBGPRINTF((OST, "->cpiParseRightSibling() parser = 0x%x, context = 0x%x, element = 0x%x\n", parser, context, element));
#endif
  
   while ((!cpiElementCompleteNext(&rc, cpiParent(&rc, element))) && (!cpiNextSibling(&rc, element)) && (pc->iCurrentElement))
      pc->iCurrentElement = parseNextItem(parser, context, pc->iCurrentElement);


#ifdef _DEBUG  
   DBGPRINTF((OST, "<-cpiParseRightSibling()\n"));
#endif

   return;
}


/******************************************************************************/
/*                                                                            */
/* Plugin Parser Implementation Function:                                     */
/*                                                                            */
/******************************************************************************/
void cpiNextParserClassName(CciParser *parser, CciContext *context, CciChar *buffer, int size)
{

#ifdef _DEBUG  
   DBGPRINTF((OST, "->cpiNextParserClassName()\n"));
#endif

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
  CciCharNCpy(buffer, constNone, size);
#else
   wcsncpy(buffer, L"none", size);
#endif

#ifdef _DEBUG  
   DBGPRINTF((OST, "<-cpiNextParserClassName()\n"));
#endif

   return;
}


/******************************************************************************/
/*                                                                            */
/* Plugin Parser Implementation Function:                                     */
/*                                                                            */
/******************************************************************************/
void cpiSetNextParserClassName(CciParser *parser, CciContext *context, CciChar *name, CciBool ParserType)
{
   PARSER_CONTEXT_ST *pc = (PARSER_CONTEXT_ST*)context;

#ifdef _DEBUG  
   DBGPRINTF((OST, "->cpiSetNextParserClassName() name = %S, ParserType = %d\n", name, ParserType));
#endif

   /* Save the name in my context */
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
   CciCharNCpy(pc->NextParserClassName, name, CciCharLen(name));
#else
   wcsncpy(pc->NextParserClassName, name, wcslen(name));
#endif
   /* Save the type in my context */
   pc->ParserType = ParserType;

#ifdef _DEBUG  
   DBGPRINTF((OST, "<-cpiSetNextParserClassName()\n"));
#endif

  return;
}

/******************************************************************************/
/*                                                                            */
/* Plugin Parser Implementation Function:                                     */
/*                                                                            */
/******************************************************************************/
CciBool cpiParserType(CciParser *parser, CciContext *context )
{

#ifdef _DEBUG  
   DBGPRINTF((OST, "<->cpiParserType(), retvalue = 1\n"));
#endif
   /* Is our parser capable of parsing MQ headers? */
   return 1; /* No, it is not */
}


/******************************************************************************/
/*                                                                            */
/* Plugin Parser Utility Function: Parse the next item                        */
/*                                                                            */
/******************************************************************************/
void *parseNextItem(CciParser *parser, CciContext *context, CciElement *element )
{
   PARSER_CONTEXT_ST *pc = (CciContext *)context;

   CciElement  *returnElement = element;
   CciElement  *newElement    = NULL;

   void  *startMarker   = NULL;

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
	const CciChar*     data;
#else
	CciChar            data[256];
#endif

   int rc    = -1;

#ifdef _DEBUG    	
   DBGPRINTF((OST, "->parseNextItem() parser = 0x%x, context = 0x%x, element = 0x%x, index = %ld\n", parser, context, element, pc->iIndex));
#endif

   if (pc->iIndex >= pc->iSize) {
#ifdef _DEBUG  
      DBGPRINTF((OST, "-->End of buffer reached...\n"));
#endif
      cpiSetElementCompletePrevious(&rc, element, 1);
      cpiSetElementCompleteNext(&rc, element, 1);
#ifdef _DEBUG  
      DBGPRINTF((OST, "-->Set completePrevious/Next on element = 0x%x\n", element));
      DBGPRINTF((OST, "<-parseNextItem() retvalue = 0\n"));
#endif
      return 0;
   }

#ifdef _DEBUG  
      DBGPRINTF((OST, "-->pc->iSize - pc->iIndex = %d, remaining %d\n", pc->iSize - pc->iIndex, ( (pc->iSize - pc->iIndex) < 79 ) ? (pc->iSize - pc->iIndex) : 79));
#endif

   if (pc->iIndex < pc->iSize) {
      /* Get a pointer to the start of the tag */
      startMarker = (char*)pc->iBuffer+(int)pc->iIndex;
      /* Create a new name element for the record */
      newElement = cpiCreateElement(&rc, parser);
      cpiSetElementType(&rc, newElement, CCI_ELEMENT_TYPE_NAME_VALUE); 
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
		cpiSetElementName(&rc, newElement, (CciChar *)constRecord);
#else
      cpiSetElementName(&rc, newElement, L"Record");
#endif
#ifdef _DEBUG  
      DBGPRINTF((OST, "-->Created new NAME element = 0x%x, type = 0x%x, index = %ld\n", newElement, CCI_ELEMENT_TYPE_NAME, pc->iIndex));
#endif          
      /* Add the element */
      cpiAddAsLastChild(&rc, element, newElement);
      cpiSetElementCompletePrevious(&rc, newElement, 1);
      cpiSetElementCompleteNext(&rc, newElement, 1);

#ifdef _DEBUG  
      DBGPRINTF((OST, "-->Added 0x%x as lastChild of 0x%x\n", newElement, element));
      DBGPRINTF((OST, "-->Set completePrevious/Next on element = 0x%x\n", newElement));
#endif

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      /* Copy the tag into a CciChar form */
      data = CciNString((char *)startMarker, ( ((pc->iSize - pc->iIndex) < 79)  ? (pc->iSize - pc->iIndex) : 79) );
#else
      /* Copy the text into a wide string */
      memset(&data, 0, sizeof(data));
      mbstowcs(data, startMarker, (unsigned int)( ((pc->iSize - pc->iIndex) < 79 ) ? (pc->iSize - pc->iIndex) : 79) );
		
#endif

      /* Set a new value element for the value */
      cpiSetElementCharacterValue(&rc, newElement, data, ( ((pc->iSize - pc->iIndex) < 79 ) ? (pc->iSize - pc->iIndex) : 79));
   
#ifdef _DEBUG  
      DBGPRINTF((OST, "-->Created new VALUE element = 0x%x, type = 0x%x, data = %S,index = %ld\n", newElement, CCI_ELEMENT_TYPE_VALUE, data, pc->iIndex));
#endif
      pc->iIndex = pc->iIndex + ( ((pc->iSize - pc->iIndex) < 79 ) ? (pc->iSize - pc->iIndex) : 79);

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
      /* Free the memory allocated in CciNString() */
      free((void *)data);
#endif

   } /* if (context->iIndex <= context->iSize) */
   
#ifdef _DEBUG          
   DBGPRINTF((OST, "<-parseNextItem() retvalue = 0x%x\n", returnElement));
#endif

   return(returnElement);
}


/******************************************************************************/
/*                                                                            */
/* Plugin Parser Implementation Function:                                     */
/*                                                                            */
/******************************************************************************/
int cpiParseBuffer(CciParser *parser, CciContext *context,int offset)
{
   PARSER_CONTEXT_ST *pc = context;

   int rc = -1;

#ifdef _DEBUG  
   DBGPRINTF((OST, "->cpiParseBuffer() parser = 0x%x, context = 0x%x, offset = %d\n", parser, context, offset));
#endif
  
   /* Get a pointer to the message buffer and save the current offset */
   pc->iBuffer = (void *)cpiBufferPointer(&rc, parser);

   if (pc->iBuffer)
      pc->iIndex = offset;
   else 
      pc->iIndex = 0;

   /* Save size of the buffer */
   pc->iSize = cpiBufferSize(&rc, parser);

   /* Prime the first byte in the stream                              */
   /* Not needed for this implementation of the parser ..             */
   /* pc->iCurrentCharacter = cpiBufferByte(&rc, parser, pc->iIndex); */

   /* Set the current element to the root element */
   pc->iCurrentElement = cpiRootElement(&rc, parser);

#ifdef _DEBUG
   DBGPRINTF((OST, "--> pc->iIndex		 = %d\n", pc->iIndex));
	DBGPRINTF((OST, "--> pc->iSize		 = %d\n", pc->iSize));
   DBGPRINTF((OST, "--> iCurrentElement = 0x%x\n", pc->iCurrentElement));
	DBGPRINTF((OST, "<-cpiParseBuffer() pc->iSize = retvalue = %d\n", pc->iSize));
#endif

//  return(pc->iIndex);
	return(pc->iSize);
}


int WriteElement(CciParser *parser, CciElement *elt)
{
   int err = 0;

   char *str = NULL;
   char sFiller[80];

   CciSize outBufLen = INITIAL_BUF_LEN;
      
   CciChar *outBuf = NULL;

   size_t Filler = 0;

#ifdef _DEBUG
   DBGPRINTF((OST, "->WriteElement element = 0x%x\n", elt));
#endif

   memset(sFiller, ' ', sizeof(sFiller));

   outBuf = (CciChar *)malloc(sizeof(CciChar) * outBufLen);

   if (elt != NULL) {
	      do {
	         cpiElementCharacterValue(&err, elt, outBuf, outBufLen);
	         if (err == CCI_BUFFER_TOO_SMALL) {
		         outBufLen *= 2;
		         outBuf = (CciChar *)realloc(outBuf, sizeof(CciChar) * outBufLen);
	         }
	      } while (err == CCI_BUFFER_TOO_SMALL);

	      if (err == 0) {
	         str = UnicodeToASCII(outBuf);
	         cpiAppendToBuffer(&err, parser, str, strlen(str));
            if ((Filler = 79 - strlen(str)) > 0 )
               cpiAppendToBuffer(&err, parser, sFiller, Filler);
	         free(str);
         }
   } /* if ((elt != NULL) && (eltName != NULL)) */

	free(outBuf);																					/* 28/07/2001 */

#ifdef _DEBUG
   DBGPRINTF((OST, "<-WriteElement retvalue = %d\n", err));
#endif

   return err;
}


int cpiWriteBuffer(CciParser *parser, CciContext *context)
{
   int err         = -1;
   int retVal      = -1;
   int recordCount = 0;

   CciElement *rootElt    = NULL;
   CciElement *curElement = NULL;

   CciChar eltName[INITIAL_BUF_LEN];

#ifdef _DEBUG
   DBGPRINTF((OST, "->cpiWriteBuffer\n"));
#endif

   rootElt = cpiRootElement(&err, parser);

   /* Output all of our children */
   curElement = cpiFirstChild(&err, rootElt);

   while (curElement != NULL && (err == 0)) {
      cpiElementName(&err, curElement, eltName, INITIAL_BUF_LEN);
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
		if (CciCharCmp(eltName, (CciChar *)constRecord) == 0) {
#else
      if (wcscmp(eltName, L"Record") == 0) {
#endif
         err = WriteElement(parser, curElement);
         recordCount++;
      }
	   curElement = cpiNextSibling(&err, curElement);
   } /* while (curElement != NULL && (err == 0)) */

   retVal = recordCount * (int)79;

#ifdef _DEBUG
   DBGPRINTF((OST, "<-cpiWriteBuffer retvalue = %d\n", retVal));
#endif
   return retVal;
}


#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
/******************************************************************************/
/*                                                                            */
/* Plugin Node Utility function:                                              */
/*                                                                            */
/*  initParserConstants                                                       */
/*                                                                            */
/*  Creates Cci constant strings from defined char * constant strings.        */
/*                                                                            */
/******************************************************************************/
void initParserConstants(void)
{
  constFFMSG001      = (CciChar*)CciString(CONST_FFMSG001);
  constParserFactory = (CciChar*)CciString(CONST_PARSER_FACTORY);
  constNone          = (CciChar*)CciString(CONST_NONE);
  constRecord        = (CciChar*)CciString(CONST_RECORD);
}
#endif


void LilFactoryExportPrefix *LilFactoryExportSuffix bipGetParserFactory()
{
  /* Declare variables */
  CciFactory*     factoryObject;
  int             rc;
  static CPI_VFT  vftable;

#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
  /* Before we proceed we need to initialise all the static constants */
  /* that may be used by the plug-in.                                 */
  initParserConstants();
#endif		
  /* Setup function table with pointers to parser implementation functions */
  memset(&vftable, 0, sizeof(vftable));
  vftable.iFpCreateContext            = cpiCreateContext;
  vftable.iFpParseBuffer              = cpiParseBuffer;
  vftable.iFpParseFirstChild          = cpiParseFirstChild;
  vftable.iFpParseLastChild           = cpiParseLastChild;
  vftable.iFpParseLeftSibling         = cpiParseLeftSibling;
  vftable.iFpParseRightSibling        = cpiParseRightSibling;
  vftable.iFpWriteBuffer              = cpiWriteBuffer;
  vftable.iFpDeleteContext            = cpiDeleteContext;
  vftable.iFpSetValue                 = 0;//cpiSetValue;
  vftable.iFpValue                    = 0;//cpiValue;
  vftable.iFpNextParserClassName      = cpiNextParserClassName;
  vftable.iFpSetNextParserClassName   = cpiSetNextParserClassName;
  vftable.iFpNextParserEncoding       = 0;//cpiNextParserEncoding;
  vftable.iFpNextParserCodedCharSetId = 0;//cpiNextParserCodedCharSetId;
  vftable.iFpIsHeaderParser           = cpiParserType;

  /* Create the parser factory for this plugin */
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
  factoryObject = cpiCreateParserFactory(&rc, constParserFactory); 
#else
  factoryObject = cpiCreateParserFactory(&rc, L"ComIbmFFMSG001ParserFactory");
#endif
  if (factoryObject) {
    /* Define the classes of message supported by the factory */
#if defined(__CCI_SOLARIS__) || defined(__CCI_HPUX__)|| defined(__CCI_LINUX__)/* 16/09/2001 */
    cpiDefineParserClass(&rc, factoryObject, constFFMSG001, &vftable);
#else
    cpiDefineParserClass(&rc, factoryObject, L"FFMSG001", &vftable);
#endif
  }
  else {
    /* Error: Unable to create parser factory */ 
  }

  /* Return address of this factory object to the broker */
  return(factoryObject);
}
